#ifndef canvas_Persistency_Common_FindOne_h
#define canvas_Persistency_Common_FindOne_h
////////////////////////////////////////////////////////////////////////
// FindOne
//
// A smart query object used as the main way of accessing associated
// objects in a one-to-one association.
//
// Given an Assns associating A with B (or B with A) (possibly with an
// associated data object D) and a source ACOLL of A objects to be
// found in the Assns, allow indexed access to the B and/or D objects
// associated with the A objects in ACOLL.
//
////////////////////////////////////
// Interface.
//////////
//
// For ease of understanding, the interface is presented here; reading
// the rest of the header is not for the faint of heart.
//
// Notes:
//
// * ProdB and Data (if needed) are the only template arguments that
// must be specified when constructing a FindOne. Any other
// items are deducible from arguments.
//
// * An attempt to create a FindOne where one of the listed
// A objects actually has multiple B objects associated with it will
// result in an exception: use FindMany{,P} instead.
//
// * The FindOne needs a source of objects of type A, an
// event and an input tag corresponding to the underlying association
// collection from which to create itself.
//
// * When constructed, the FindOne will obtain and
// interrogate the correct Assns and provide access to the B (and/or D)
// object(s) associated with each supplied A object in the order in
// which the A objects were specified.
//
// * If the specified A does not have an associated B or D then the
// cet::maybe_ref will be invalid.
//
// * If the required association collection has an extra data object D
// with each association then it *must* be specified as a template
// argument, even if it is not relevant to the current query.
//
// * *All* indexed accessors (at(), data(), get()) are bounds-checked.
//
// Useful typedefs.
//
// typedef ProdB assoc_t;
// typedef Data data_t;
// typedef typename std::vector<assoc_t const *>::size_type size_type; // FindOne
// typedef typename std::vector<art::Ptr<assoc_t> >::size_type size_type; // FindOneP
//
// Constructors.
//
// // From Handle or ValidHandle to collection of A.
// FindOne<ProdB>(HANDLE<ProdAColl> const &,
//                EVENT const &,
//                InputTag const &);
// FindOne<ProdB, Data>(HANDLE<ProdAColl> const &,
//                      EVENT const &,
//                      InputTag const &);
//
// // From sequence of pointer to A (including View<A>).
// FindOne<ProdB>(View<ProdA> const &,
//                EVENT const &,
//                InputTag const &);
// FindOne<ProdB, Data>(View<ProdA> const &,
//                      EVENT const &,
//                      InputTag const &);
//
// // From arbitrary sequence of Ptr<A>.
// FindOne<ProdB>(PtrProdAColl const &,
//                EVENT const &,
//                InputTag const &);
// FindOne<ProdB, Data>(PtrProdAColl const &,
//                      EVENT const &,
//                      InputTag const &);
//
// // From an initializer list of Ptr<A>.
// FindOne<ProdB>(<brace-enclosed initializer list>,
//                EVENT const &,
//                InputTag const &);
// FindOne<ProdB, Data>(<brace-enclosed initializer list>,
//                      EVENT const &,
//                      InputTag const &);
//
// Modifiers.
//
// <NONE>.
//
// Accessors.
//
// size_type size() const;
// cet::maybe_ref<assoc_t const> at(size_type) const; // FindOne
// art::Ptr<assoc_t> const & at(size_type) const; // FindOneP
// cet::maybe_ref<data_t const> data(size_type) const;
// void get(size_type,
//          cet::maybe_ref<assoc_t const> &) const; // FindOne
// void get(size_type,
//          art::Ptr<assoc_t> &) const; // FindOneP
// void get(size_type,
//          cet::maybe_ref<assoc_t const> &,
//          cet::maybe_ref<data_t const> &)
//   const; // *Must* be used for FindOne<ProdB, Data>.
// void get(size_type,
//          art::Ptr<assoc_t> &,
//          cet::maybe_ref<data_t const> &)
//   const; // *Must* be used for FindOneP<ProdB, Data>.
//
// Comparison operations.
//
// bool operator == (FindOne const & other) const;
//
////////////////////////////////////////////////////////////////////////
#include "canvas/Persistency/Common/detail/IPRHelper.h"
#include "canvas/Persistency/Common/detail/is_handle.h"
#include "canvas/Persistency/Common/Assns.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Utilities/InputTag.h"
#include "cetlib/maybe_ref.h"

#include <initializer_list>
#include <vector>

/* #undef ART_IPR_BY_PTR */

namespace art {
  // General template
  template <typename ProdB, typename DATA = void>
  class FindOne;

  // Specialization.
  template <typename ProdB>
  class FindOne<ProdB, void>;
}

////////////////////////////////////////////////////////////////////////
// Implementation of the specialization.
template <typename ProdB>
class art::FindOne<ProdB, void> {
public:

  typedef ProdB assoc_t;

#ifdef ART_IPR_BY_PTR
  typedef std::vector<Ptr<assoc_t> > bColl_t;
#else
  typedef std::vector<ProdB const *> bColl_t;
#endif
  typedef typename bColl_t::value_type value_type;
  typedef typename bColl_t::size_type size_type;
  typedef typename bColl_t::difference_type difference_type;
  typedef typename bColl_t::const_reference const_reference;
  typedef typename bColl_t::reference reference;

  template <typename HANDLE, typename EVENT>
    FindOne(HANDLE const & aCollection,
                         EVENT const & e,
                         InputTag const & tag,
                         typename std::enable_if<detail::is_handle<HANDLE>::value>::type * = 0);

  template <typename ProdAColl, typename EVENT>
    FindOne(ProdAColl const & view,
                         EVENT const & e,
                         InputTag const & tag,
                         typename std::enable_if<std::is_pointer<typename ProdAColl::value_type>::value, void>::type * = 0);

  template <typename PtrProdAColl, typename EVENT>
    FindOne(PtrProdAColl const & aPtrColl,
                         EVENT const & e,
                         InputTag const & tag,
                         typename std::enable_if<std::is_same<typename PtrProdAColl::value_type,
                         art::Ptr<typename PtrProdAColl::value_type::value_type> >::value>::type * = 0);

  template <typename ProdA, typename EVENT>
    FindOne(std::initializer_list<Ptr<ProdA> > const & ptrs,
                         EVENT const & e,
                         InputTag const & tag);

  // Is this a valid query (did we find an Assns)?
  bool isValid() const;

  // Number of query results
  size_type size() const;

  // Associated item by index (bounds-checked).
#ifdef ART_IPR_BY_PTR
  Ptr<assoc_t> const & at(size_type i) const;
  void get(size_type i, Ptr<assoc_t> & item) const;
#else
  cet::maybe_ref<assoc_t const> at(size_type i) const;
  void get(size_type i, cet::maybe_ref<assoc_t const> & item) const;
#endif
  bool operator == (FindOne<ProdB, void> const & other) const;

protected:
  FindOne() : bCollection_() { }
  bColl_t & bCollection() { return bCollection_; }

  void setStoredException(std::shared_ptr<art::Exception const> && e);
  void throwIfInvalid() const;

private:
  bColl_t bCollection_;
  std::shared_ptr<art::Exception const> storedException_;
};

template <typename ProdB, typename Data>
class art::FindOne : private art::FindOne<ProdB, void> {
private:
  typedef FindOne<ProdB, void> base;
public:
  typedef std::vector<Data const *> dataColl_t;
  typedef typename base::value_type value_type;
  typedef typename base::size_type size_type;
  typedef typename base::difference_type difference_type;
  typedef typename base::const_reference const_reference;
  typedef typename base::reference reference;
  typedef typename base::assoc_t assoc_t;

  typedef Data const * data_const_pointer;
  typedef Data const & data_const_reference;
  typedef Data & data_reference;

  typedef Data data_t;

  template <typename HANDLE, typename EVENT>
    FindOne(HANDLE const & aCollection,
                         EVENT const & e,
                         InputTag const & tag,
                         typename std::enable_if<detail::is_handle<HANDLE>::value>::type * = 0);

  template <typename ProdAColl, typename EVENT>
    FindOne(ProdAColl const & view,
                         EVENT const & e,
                         InputTag const & tag,
                         typename std::enable_if<std::is_pointer<typename ProdAColl::value_type>::value, void>::type * = 0);


  template <typename PtrProdAColl, typename EVENT>
    FindOne(PtrProdAColl const & aPtrColl,
                         EVENT const & e,
                         InputTag const & tag,
                         typename std::enable_if<std::is_same<typename PtrProdAColl::value_type,
                         art::Ptr<typename PtrProdAColl::value_type::value_type> >::value>::type * = 0);

  template <typename ProdA, typename EVENT>
    FindOne(std::initializer_list<Ptr<ProdA> > const & ptrs,
                         EVENT const & e,
                         InputTag const & tag);

  using base::size;
  using base::isValid;
  using base::at;
  using base::get;

  // Association extra-data object by index (bounds-checked).
  cet::maybe_ref<Data const> data(size_type i) const;

  // Associated item and extra-data object by index (bounds-checked).
  void get(size_type i,
#ifdef ART_IPR_BY_PTR
           art::Ptr<assoc_t> & item,
#else
           cet::maybe_ref<assoc_t const> & item,
#endif
           cet::maybe_ref<Data const> & data) const;

  bool operator == (FindOne<ProdB, Data> const & other) const;

private:
  dataColl_t dataCollection_;
};

////////////////////////////////////////////////////////////////////////
// Base class implementation.
template <typename ProdB>
template <typename HANDLE, typename EVENT>
art::FindOne<ProdB, void>::
FindOne(HANDLE const & aCollection,
                     EVENT const & e,
                     InputTag const & tag,
                     typename std::enable_if<detail::is_handle<HANDLE>::value>::type *)
  :
  bCollection_(),
  storedException_()
{
  typedef typename HANDLE::element_type::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, void, void, EVENT> finder(e, tag);
  storedException_ =
    finder(*aCollection, bCollection_);
}

template <typename ProdB>
template <typename ProdAColl, typename EVENT>
art::FindOne<ProdB, void>::
FindOne(ProdAColl const & view,
                     EVENT const & e,
                     InputTag const & tag,
                     typename std::enable_if<std::is_pointer<typename ProdAColl::value_type>::value, void>::type *)
  :
  bCollection_(),
  storedException_()
{
  typedef typename std::remove_const<typename std::remove_pointer<typename ProdAColl::value_type>::type>::type ProdA;
  detail::IPRHelper<ProdA, ProdB, void, void, EVENT> finder(e, tag);
  storedException_ =
    finder(view, bCollection_);
}

template <typename ProdB>
template <typename PtrProdAColl, typename EVENT>
art::FindOne<ProdB, void>::
FindOne(PtrProdAColl const & aPtrColl,
                     EVENT const & e,
                     InputTag const & tag,
                     typename std::enable_if<std::is_same<typename PtrProdAColl::value_type,
                     art::Ptr<typename PtrProdAColl::value_type::value_type> >::value>::type *)
  :
  bCollection_(),
  storedException_()
{
  typedef typename PtrProdAColl::value_type::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, void, void, EVENT> finder(e, tag);
  storedException_ =
    finder(aPtrColl, bCollection_);
}

template <typename ProdB>
template <typename ProdA, typename EVENT>
art::FindOne<ProdB, void>::
FindOne(std::initializer_list<Ptr<ProdA> > const & ptrs,
                     EVENT const & e,
                     InputTag const & tag)
  :
  bCollection_(),
  storedException_()
{
  detail::IPRHelper<ProdA, ProdB, void, void, EVENT> finder(e, tag);
  storedException_ =
    finder(ptrs, bCollection_);
}

template <typename ProdB>
inline
auto
art::FindOne<ProdB, void>::
size() const
-> size_type
{
  throwIfInvalid();
  return bCollection_.size();
}

template <typename ProdB>
inline
bool
art::FindOne<ProdB, void>::
isValid() const
{
  return (storedException_.get() == nullptr);
}

template <typename ProdB>
inline
auto
art::FindOne<ProdB, void>::
at(size_type i) const
->
#ifdef ART_IPR_BY_PTR
art::Ptr<assoc_t> const &
#else
cet::maybe_ref<assoc_t const>
#endif
{
  throwIfInvalid();
#ifdef ART_IPR_BY_PTR
  return bCollection_.at(i);
#else
  return cet::maybe_ref<assoc_t const>(*bCollection_.at(i));
#endif
}

template <typename ProdB>
inline
void
art::FindOne<ProdB, void>::
get(size_type i,
#ifdef ART_IPR_BY_PTR
    Ptr<assoc_t> & item
#else
    cet::maybe_ref<assoc_t const> & item
#endif
   ) const
{
  throwIfInvalid();
#ifdef ART_IPR_BY_PTR
  item = bCollection_.at(i);
#else
  item.reseat(*bCollection_.at(i));
#endif
}

template <typename ProdB>
inline
bool
art::FindOne<ProdB, void>::
operator == (FindOne<ProdB, void> const & other) const
{
  throwIfInvalid();
  return bCollection_ == other.bCollection_;
}

template <typename ProdB>
inline
void
art::FindOne<ProdB, void>::
setStoredException(std::shared_ptr<art::Exception const> && e)
{
  storedException_ = std::move(e);
}

template <typename ProdB>
inline
void
art::FindOne<ProdB, void>::
throwIfInvalid() const
{
  if (!isValid()) {
    throw Exception(errors::LogicError, "Invalid FindOne", *storedException_)
      << "Attempt to use a FindOne where the underlying art::Assns product was not found.";
  }
}

////////////////////////////////////////////////////////////////////////
// Derived class implementation.
template <typename ProdB, typename Data>
template <typename HANDLE, typename EVENT>
art::FindOne<ProdB, Data>::
FindOne(HANDLE const & aCollection,
                     EVENT const & e,
                     InputTag const & tag,
                     typename std::enable_if<detail::is_handle<HANDLE>::value>::type *)
    :
    base(),
    dataCollection_()
{
  typedef typename HANDLE::element_type::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t, EVENT> finder(e, tag);
  base::setStoredException(finder(*aCollection, base::bCollection(), dataCollection_));
}

template <typename ProdB, typename Data>
template <typename ProdAColl, typename EVENT>
art::FindOne<ProdB, Data>::
FindOne(ProdAColl const & view,
                     EVENT const & e,
                     InputTag const & tag,
                         typename std::enable_if<std::is_pointer<typename ProdAColl::value_type>::value, void>::type *)
    :
    base(),
    dataCollection_()
{
  typedef typename std::remove_const<typename std::remove_pointer<typename ProdAColl::value_type>::type>::type ProdA;
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t, EVENT> finder(e, tag);
  base::setStoredException(finder(view, base::bCollection(), dataCollection_));
}

template <typename ProdB, typename Data>
template <typename PtrProdAColl, typename EVENT>
art::FindOne<ProdB, Data>::
FindOne(PtrProdAColl const & aPtrColl,
                     EVENT const & e,
                     InputTag const & tag,
                     typename std::enable_if<std::is_same<typename PtrProdAColl::value_type,
                     art::Ptr<typename PtrProdAColl::value_type::value_type> >::value>::type *)
    :
    base(),
    dataCollection_()
{
  typedef typename PtrProdAColl::value_type::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t, EVENT> finder(e, tag);
  base::setStoredException(finder(aPtrColl, base::bCollection(), dataCollection_));
}

template <typename ProdB, typename Data>
template <typename ProdA, typename EVENT>
art::FindOne<ProdB, Data>::
FindOne(std::initializer_list<Ptr<ProdA> > const & ptrs,
                     EVENT const & e,
                     InputTag const & tag)
    :
    base(),
    dataCollection_()
{
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t, EVENT> finder(e, tag);
  base::setStoredException(finder(ptrs, base::bCollection(), dataCollection_));
}

template <typename ProdB, typename Data>
inline
cet::maybe_ref<Data const>
art::FindOne<ProdB, Data>::
data(size_type i) const
{
  base::throwIfInvalid();
  return cet::maybe_ref<Data const>(*dataCollection_.at(i));
}

template <typename ProdB, typename Data>
inline
void
art::FindOne<ProdB, Data>::
get(size_type i,
#ifdef ART_IPR_BY_PTR
    Ptr<assoc_t> & item,
#else
    cet::maybe_ref<assoc_t const> & item,
#endif
    cet::maybe_ref<Data const> & data) const
{
  base::get(i, item); // Will check validity.
  data.reseat(*dataCollection_.at(i));
}

template <typename ProdB, typename Data>
inline
bool
art::FindOne<ProdB, Data>::
operator == (FindOne<ProdB, Data> const & other) const
{
  return this->base::operator==(other) && // Will check validity.
         dataCollection_ == other.dataCollection_;
}

#undef ART_IPR_BY_PTR

#endif /* canvas_Persistency_Common_FindOne_h */

// Local Variables:
// mode: c++
// End:
